@@ -48,6 +48,7 @@ seafile_crypt_new (int version, unsigned char *key, unsigned char *iv)
 
 int
 seafile_derive_key (const char *data_in, int in_len, int version,
+                    const char *repo_salt,
                     unsigned char *key, unsigned char *iv)
 {
 #ifdef USE_GPL_CRYPTO
@@ -62,7 +63,23 @@ seafile_derive_key (const char *data_in, int in_len, int version,
 
     return 0;
 #else
-    if (version == 2) {
+    if (version == 3) {
+        unsigned char repo_salt_bin[32];
+
+        hex_to_rawdata (repo_salt, repo_salt_bin, 32);
+
+        PKCS5_PBKDF2_HMAC (data_in, in_len,
+                           repo_salt_bin, sizeof(repo_salt_bin),
+                           KEYGEN_ITERATION2,
+                           EVP_sha256(),
+                           32, key);
+        PKCS5_PBKDF2_HMAC ((char *)key, 32,
+                           repo_salt_bin, sizeof(repo_salt_bin),
+                           10,
+                           EVP_sha256(),
+                           16, iv);
+        return 0;
+    } else if (version == 2) {
         PKCS5_PBKDF2_HMAC (data_in, in_len,
                            salt, sizeof(salt),
                            KEYGEN_ITERATION2,
@@ -99,7 +116,26 @@ seafile_derive_key (const char *data_in, int in_len, int version,
 }
 
 int
-seafile_generate_random_key (const char *passwd, char *random_key)
+seafile_generate_repo_salt (char *repo_salt)
+{
+    unsigned char repo_salt_bin[32];
+
+    int rc = RAND_bytes (repo_salt_bin, sizeof(repo_salt_bin));
+    if (rc != 1) {
+        seaf_warning ("Failed to generate salt for repo encryption.\n");
+        return -1;
+    }
+
+    rawdata_to_hex (repo_salt_bin, repo_salt, 32);
+
+    return 0;
+}
+
+int
+seafile_generate_random_key (const char *passwd,
+                             int version,
+                             const char *repo_salt,
+                             char *random_key)
 {
     SeafileCrypt *crypt;
     unsigned char secret_key[32], *rand_key;
@@ -113,15 +149,14 @@ seafile_generate_random_key (const char *passwd, char *random_key)
     }
 #else
     if (RAND_bytes (secret_key, sizeof(secret_key)) != 1) {
-        seaf_warning ("Failed to generate secret key for repo encryption "
-                      "with RAND_bytes(), use RAND_pseudo_bytes().\n");
-        RAND_pseudo_bytes (secret_key, sizeof(secret_key));
+        seaf_warning ("Failed to generate secret key for repo encryption.\n");
+        return -1;
     }
 #endif
 
-    seafile_derive_key (passwd, strlen(passwd), 2, key, iv);
+    seafile_derive_key (passwd, strlen(passwd), version, repo_salt, key, iv);
 
-    crypt = seafile_crypt_new (2, key, iv);
+    crypt = seafile_crypt_new (version, key, iv);
 
     seafile_encrypt ((char **)&rand_key, &outlen,
                      (char *)secret_key, sizeof(secret_key), crypt);
@@ -136,7 +171,9 @@ seafile_generate_random_key (const char *passwd, char *random_key)
 
 void
 seafile_generate_magic (int version, const char *repo_id,
-                        const char *passwd, char *magic)
+                        const char *passwd,
+                        const char *repo_salt,
+                        char *magic)
 {
     GString *buf = g_string_new (NULL);
     unsigned char key[32], iv[16];
@@ -147,7 +184,7 @@ seafile_generate_magic (int version, const char *repo_id,
      */
     g_string_append_printf (buf, "%s%s", repo_id, passwd);
 
-    seafile_derive_key (buf->str, buf->len, version, key, iv);
+    seafile_derive_key (buf->str, buf->len, version, repo_salt, key, iv);
 
     g_string_free (buf, TRUE);
     rawdata_to_hex (key, magic, 32);
@@ -157,13 +194,14 @@ int
 seafile_verify_repo_passwd (const char *repo_id,
                             const char *passwd,
                             const char *magic,
-                            int version)
+                            int version,
+                            const char *repo_salt)
 {
     GString *buf = g_string_new (NULL);
     unsigned char key[32], iv[16];
     char hex[65];
 
-    if (version != 1 && version != 2) {
+    if (version != 1 && version != 2 && version != 3) {
         seaf_warning ("Unsupported enc_version %d.\n", version);
         return -1;
     }
@@ -171,10 +209,11 @@ seafile_verify_repo_passwd (const char *repo_id,
     /* Recompute the magic and compare it with the one comes with the repo. */
     g_string_append_printf (buf, "%s%s", repo_id, passwd);
 
-    seafile_derive_key (buf->str, buf->len, version, key, iv);
+    seafile_derive_key (buf->str, buf->len, version, repo_salt, key, iv);
 
     g_string_free (buf, TRUE);
-    if (version == 2)
+
+    if (version >= 2)
         rawdata_to_hex (key, hex, 32);
     else
         rawdata_to_hex (key, hex, 16);
@@ -188,17 +227,18 @@ seafile_verify_repo_passwd (const char *repo_id,
 int
 seafile_decrypt_repo_enc_key (int enc_version,
                               const char *passwd, const char *random_key,
+                              const char *repo_salt,
                               unsigned char *key_out, unsigned char *iv_out)
 {
     unsigned char key[32], iv[16];
 
-    seafile_derive_key (passwd, strlen(passwd), enc_version, key, iv);
+    seafile_derive_key (passwd, strlen(passwd), enc_version, repo_salt, key, iv);
 
     if (enc_version == 1) {
         memcpy (key_out, key, 16);
         memcpy (iv_out, iv, 16);
         return 0;
-    } else if (enc_version == 2) {
+    } else if (enc_version >= 2) {
         unsigned char enc_random_key[48], *dec_random_key;
         int outlen;
         SeafileCrypt *crypt;
@@ -221,7 +261,8 @@ seafile_decrypt_repo_enc_key (int enc_version,
         g_free (crypt);
 
         seafile_derive_key ((char *)dec_random_key, 32, enc_version,
-                                  key, iv);
+                            repo_salt,
+                            key, iv);
         memcpy (key_out, key, 32);
         memcpy (iv_out, iv, 16);
 
@@ -234,7 +275,8 @@ seafile_decrypt_repo_enc_key (int enc_version,
 
 int
 seafile_update_random_key (const char *old_passwd, const char *old_random_key,
-                           const char *new_passwd, char *new_random_key)
+                           const char *new_passwd, char *new_random_key,
+                           int enc_version, const char *repo_salt)
 {
     unsigned char key[32], iv[16];
     unsigned char random_key_raw[48], *secret_key, *new_random_key_raw;
@@ -242,11 +284,12 @@ seafile_update_random_key (const char *old_passwd, const char *old_random_key,
     SeafileCrypt *crypt;
 
     /* First, use old_passwd to decrypt secret key from old_random_key. */
-    seafile_derive_key (old_passwd, strlen(old_passwd), 2, key, iv);
+    seafile_derive_key (old_passwd, strlen(old_passwd), enc_version,
+                        repo_salt, key, iv);
 
     hex_to_rawdata (old_random_key, random_key_raw, 48);
 
-    crypt = seafile_crypt_new (2, key, iv);
+    crypt = seafile_crypt_new (enc_version, key, iv);
     if (seafile_decrypt ((char **)&secret_key, &secret_key_len,
                          (char *)random_key_raw, 48,
                          crypt) < 0) {
@@ -257,9 +300,9 @@ seafile_update_random_key (const char *old_passwd, const char *old_random_key,
     g_free (crypt);
 
     /* Second, use new_passwd to encrypt secret key. */
-    seafile_derive_key (new_passwd, strlen(new_passwd), 2, key, iv);
-
-    crypt = seafile_crypt_new (2, key, iv);
+    seafile_derive_key (new_passwd, strlen(new_passwd), enc_version,
+                        repo_salt, key, iv);
+    crypt = seafile_crypt_new (enc_version, key, iv);
 
     seafile_encrypt ((char **)&new_random_key_raw, &random_key_len,
                      (char *)secret_key, secret_key_len, crypt);
