@@ -42,6 +42,7 @@ struct _SeafRepo {
     gchar      *category;       /* not used yet */
     gboolean    encrypted;
     int         enc_version;
+    gchar       salt[65];
     gchar       magic[65];       /* hash(repo_id + passwd), key stretched. */
     gchar       random_key[97];  /* key length is 48 after encryption */
     gboolean    no_local_history;
@@ -413,6 +414,7 @@ seaf_repo_manager_is_path_writable (SeafRepoManager *mgr,
 #define SYNC_ERROR_ID_PATH_INVALID_CHARACTER 6
 #define SYNC_ERROR_ID_FOLDER_PERM_DENIED 7
 #define SYNC_ERROR_ID_PERM_NOT_SYNCABLE 8
+#define SYNC_ERROR_ID_UPDATE_TO_READ_ONLY_REPO 9
 
 GList *
 seaf_repo_manager_get_file_sync_errors (SeafRepoManager *mgr, int offset, int limit);
